import time
import json
import csv
import gzip
import logging
import io
from collections import namedtuple
from ..common.util import set_dataframe_types
import pandas as pd
from urllib.parse import quote as quote_uri

from .Upload import Upload
from .Variable import Variable
from ..common.api_request import make_request, make_paginated_request, make_rows_request


class Table:
    def __init__(
        self,
        name,
        *,
        sample=False,
        dataset=None,
        project=None,
        properties=None,
    ):
        parent = dataset or project
        owner = parent.user or parent.organization
        sample_string = ":sample" if sample else ""
        version_string = f":{dataset.version}" if dataset else ""
        self.name = name
        self.dataset = dataset
        self.project = project
        self.identifier = (
            f"{owner.name}.{parent.name}{version_string}.{self.name}{sample_string}"
        )
        self.uri = f"/tables/{quote_uri(self.identifier)}"
        self.properties = properties

    def __getitem__(self, key):
        return self.properties[key]

    def __str__(self):
        return json.dumps(self.properties, indent=2)

    def exists(self):
        try:
            make_request(method="GET", path=self.uri)
        except Exception as err:
            if err.args[0]["status"] != 404:
                raise err
            return False

    def get(self):
        self.properties = make_request(method="GET", path=self.uri)
        self.uri = self.properties["uri"]
        return self

    def create(self, *, description=None, upload_merge_strategy="append"):
        response = make_request(
            method="POST",
            path=f"{self.dataset.uri}/tables",
            payload={
                "name": self.name,
                "description": description,
                "uploadMergeStrategy": upload_merge_strategy,
            },
        )
        self.properties = response
        self.uri = self.properties["uri"]
        return self

    def update(self, *, name=None, description=None, upload_merge_strategy=None):
        payload = {}
        if name:
            payload["name"] = name
        if upload_merge_strategy:
            payload["mergeStrategy"] = upload_merge_strategy
        if description is not None:
            payload["description"] = description

        response = make_request(
            method="PATCH",
            path=f"{self.uri}",
            payload=payload,
        )
        self.properties = response
        return

    def delete(self):
        make_request(
            method="DELETE",
            path=self.uri,
        )
        return

    def upload(self, *, name, data, type, remove_on_fail=True):
        response = make_request(
            method="POST",
            path=f"{self.uri}/uploads",
            payload={"name": name, "type": type},
        )
        upload = Upload(uri=response["uri"])
        try:
            upload.upload_file(data)
            while True:
                time.sleep(2)
                upload.get()
                if upload["status"] == "completed":
                    break
                elif upload["status"] == "failed":
                    raise Exception(upload["errorMessage"])
                else:
                    logging.debug("Upload is still in progress...")
        except Exception as e:
            if remove_on_fail:
                print("An error occurred. Deleting upload.")
                upload.delete()
            raise e

        return upload

    def list_uploads(self, *, max_results=None):
        uploads = make_paginated_request(
            path=f"{self.uri}/uploads", max_results=max_results
        )
        return [Upload(upload) for upload in uploads]

    def list_variables(self, *, max_results=None):
        variables = make_paginated_request(
            path=f"{self.uri}/variables", page_size=1000, max_results=max_results
        )
        return [Variable(variable) for variable in variables]

    def list_rows(self, max_results=None, *, limit=None, variables=None):
        # TODO: limit is deprectated and should ultimately be removed
        if limit and max_results is None:
            max_results = limit

        # This allows us to persist casing of the passed variable names
        original_variable_names = variables
        all_variables = make_paginated_request(path=f"{self.uri}/variables")

        if variables is None:
            variables_list = all_variables
            original_variable_names = [variable["name"] for variable in all_variables]
        else:
            lower_variable_names = [variable.lower() for variable in variables]
            variables_list = list(
                filter(
                    lambda variable: variable["name"].lower() in lower_variable_names,
                    all_variables,
                )
            )
            variables_list.sort(
                key=lambda variable: lower_variable_names.index(
                    variable["name"].lower()
                )
            )

        Row = namedtuple(
            "Row",
            original_variable_names,
        )

        if not self.properties:
            self.get()

        max_results = (
            min(max_results, int(self.properties["numRows"]))
            if max_results is not None
            else self.properties["numRows"]
        )
        res = make_rows_request(
            uri=self.uri,
            max_results=max_results,
            query={
                "selectedVariables": ",".join(
                    map(lambda variable: variable["name"], variables_list)
                ),
            },
        )

        fd = res.raw
        if res.headers.get("content-encoding") == "gzip":
            fd = gzip.GzipFile(fileobj=fd, mode="r")

        reader = csv.reader(io.TextIOWrapper(fd))

        return [Row(*row) for row in reader]

    def to_dataframe(self, max_results=None, *, limit=None, variables=None):
        original_variable_names = variables
        all_variables = make_paginated_request(path=f"{self.uri}/variables")

        if variables is None:
            variables_list = all_variables
            original_variable_names = [variable["name"] for variable in all_variables]
        else:
            lower_variable_names = [variable.lower() for variable in variables]
            variables_list = list(
                filter(
                    lambda variable: variable["name"].lower() in lower_variable_names,
                    all_variables,
                )
            )
            variables_list.sort(
                key=lambda variable: lower_variable_names.index(
                    variable["name"].lower()
                )
            )

        # TODO: limit is deprectated and should ultimately be removed
        if limit and max_results is None:
            max_results = limit

        if not self.properties:
            self.get()

        max_results = (
            min(max_results, int(self.properties["numRows"]))
            if max_results is not None
            else self.properties["numRows"]
        )
        res = make_rows_request(
            uri=self.uri,
            max_results=max_results,
            query={
                "selectedVariables": ",".join(
                    map(lambda variable: variable["name"], variables_list)
                ),
            },
        )

        df = pd.read_csv(
            res.raw,
            dtype="string",
            names=original_variable_names,
            compression="gzip"
            if res.headers.get("content-encoding") == "gzip"
            else None,
        )

        if variables is None:
            return set_dataframe_types(df, variables_list)
        else:
            return set_dataframe_types(
                df,
                map(
                    lambda variable, variable_name: {
                        "name": variable_name,
                        "type": variable["type"],
                    },
                    variables_list,
                    variables,
                ),
            )



import pandas as pd


def set_dataframe_types(df, variables):
    for variable in variables:
        name = variable["name"]
        type = variable["type"]

        if type == "integer":
            df[name] = pd.to_numeric(df[name])
            # df[name] = df[name].astype("Int64")
        elif type == "float":
            df[name] = df[name].astype("float64")
            # df[name] = df[name].astype("Float64")
        elif type == "date":
            df[name] = pd.to_datetime(df[name], errors="coerce")
        elif type == "dateTime":
            df[name] = pd.to_datetime(df[name], errors="coerce")
        elif type == "time":
            df[name] = pd.to_timedelta(df[name], errors="coerce")
        elif type == "boolean":
            df.replace(to_replace=["true", "false"], value=[True, False], inplace=True)
            df[name] = df[name].astype("boolean")
        else:
            df[name] = df[name].astype("string")

    return df



from ..common.api_request import make_request, make_rows_request
from ..common.util import set_dataframe_types
import json
import gzip
import csv
import io
import time
from collections import namedtuple
import pandas as pd


class Query:
    def __init__(self, query, *, default_project=None, default_dataset=None):
        self.properties = make_request(
            method="post",
            path="/queries",
            payload={
                "query": query,
                "defaultProject": default_project if default_project else None,
                "defaultDataset": default_dataset if default_dataset else None,
            },
        )
        self.uri = f"/queries/{self.properties['id']}"

    def __getitem__(self, key):
        return self.properties[key]

    def __str__(self):
        return json.dumps(self.properties, indent=2)

    def get(self):
        self.properties = make_request(method="GET", path=self.uri)
        return self

    def list_rows(self, max_results=None, *, limit=None):
        if limit and max_results is None:
            max_results = limit

        self._wait_for_finish()
        variables = self.properties["outputSchema"]
        Row = namedtuple(
            "Row",
            [variable["name"] for variable in variables],
        )

        max_results = (
            min(max_results, int(self.properties["outputNumRows"]))
            if max_results is not None
            else self.properties["outputNumRows"]
        )

        res = make_rows_request(uri=self.uri, max_results=max_results)
        fd = res.raw
        if res.headers.get("content-encoding") == "gzip":
            fd = gzip.GzipFile(fileobj=fd, mode="r")

        reader = csv.reader(io.TextIOWrapper(fd))

        return [Row(*row) for row in reader]

    def to_dataframe(self, max_results=None, *, limit=None):
        if limit and max_results is None:
            max_results = limit

        self._wait_for_finish()
        variables = self.properties["outputSchema"]

        max_results = (
            min(max_results, int(self.properties["outputNumRows"]))
            if max_results is not None
            else self.properties["outputNumRows"]
        )

        res = make_rows_request(uri=self.uri, max_results=max_results)

        df = pd.read_csv(
            res.raw,
            dtype="string",
            names=[variable["name"] for variable in variables],
            compression="gzip"
            if res.headers.get("content-encoding") == "gzip"
            else None,
        )

        return set_dataframe_types(df, variables)

    def _wait_for_finish(self):
        while True:
            if self.properties["status"] == "completed":
                break
            elif self.properties["status"] == "failed":
                raise Exception(
                    f"Query job failed with message: {self.properties['errorMessage']}"
                )
            elif self.properties["status"] == "cancelled":
                raise Exception(f"Query job was cancelled")
            else:
                time.sleep(2)
                self.get()



import requests
import logging
import os
import json
from .auth import get_auth_token
from urllib3.exceptions import InsecureRequestWarning

# Suppress only the single warning from urllib3 needed.
requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)


def make_request(
    *,
    method,
    path,
    query=None,
    payload=None,
    parse_payload=True,
    parse_response=True,
    stream=False,
):
    api_endpoint = __get_api_endpoint()
    verify_ssl = (
        False
        if api_endpoint.find("https://localhost", 0) == 0
        or os.getenv("REDIVIS_ENV") == "development"
        or os.getenv("REDIVIS_ENV") == "test"
        or os.getenv("REDIVIS_ENV") == "staging"
        else True
    )
    method = method.lower()
    url = f"{api_endpoint}{path}"

    headers = {"Authorization": f"Bearer {get_auth_token()}"}

    logging.debug(f"Making API '{method}' request to '{url}'")

    if parse_payload and payload:
        payload = json.dumps(payload)

    r = getattr(requests, method)(
        url,
        headers=headers,
        params=query,
        verify=verify_ssl,
        data=payload,
        stream=stream,
    )

    response_json = {}
    try:
        if r.status_code >= 400 or (parse_response and r.text != "OK"):
            response_json = r.json()
    except Exception:
        raise Exception(r.text)

    if r.status_code >= 400:
        raise Exception(response_json["error"])
    elif parse_response:
        return response_json
    else:
        return r


def make_paginated_request(
    *, path, query={}, page_size=100, max_results=None, parse_response=True
):
    logging.debug(f"Making paginated API request to '{path}'")

    page = 0
    results = []
    next_page_token = None

    while True:
        if max_results is not None and len(results) >= max_results:
            break

        response = make_request(
            method="get",
            path=path,
            parse_response=True,
            query={
                **query,
                **{
                    "pageToken": next_page_token,
                    "maxResults": page_size
                    if max_results is None or (page + 1) * page_size < max_results
                    else max_results - page * page_size,
                },
            },
        )
        page += 1
        results += response["results"]
        next_page_token = response["nextPageToken"]
        if not next_page_token:
            break

    return results


def make_rows_request(*, uri, max_results, query={}):
    res = make_request(
        method="get",
        path=f"{uri}/rows",
        parse_response=False,
        stream=True,
        query={
            **query,
            **{"maxResults": max_results, "format": "csv"},
        },
    )

    return res


def __get_api_endpoint():
    return (
        "https://redivis.com/api/v1"
        if os.getenv("REDIVIS_API_ENDPOINT") is None
        else os.getenv("REDIVIS_API_ENDPOINT")
    )



import os


def get_auth_token():
    if os.getenv("REDIVIS_API_TOKEN") is None:
        raise EnvironmentError(
            "The environment variable REDIVIS_API_TOKEN must be set."
        )
    return os.environ["REDIVIS_API_TOKEN"]



import json
import time
import re
import math
import os
import requests
from ..common.api_request import make_request, make_paginated_request

# 8MB
MAX_CHUNK_SIZE = 2 ** 23


class Upload:
    def __init__(self, uri, properties=None):
        # TODO: reconsider this scheme
        self.uri = uri
        self.properties = properties

    def __getitem__(self, key):
        return self.properties[key]

    def __str__(self):
        return json.dumps(self.properties, indent=2)

    def get(self):
        self.properties = make_request(
            method="GET",
            path=self.uri,
        )

    def delete(self):
        make_request(
            method="DELETE",
            path=self.uri,
        )

    def upload_file(self, data):
        start_byte = 0
        retry_count = 0
        chunk_size = MAX_CHUNK_SIZE
        is_file = True if hasattr(data, "read") else False
        file_size = os.stat(data.name).st_size if is_file else len(data)

        res = make_request(
            path=f"{self.uri}/resumableUri",
            method="POST",
            payload={"size": file_size},
        )
        resumable_uri = res["uri"]

        while start_byte < file_size:
            end_byte = min(start_byte + chunk_size - 1, file_size - 1)
            if is_file:
                data.seek(start_byte)
                chunk = data.read(end_byte - start_byte + 1)
            else:
                chunk = data[start_byte:end_byte]

            try:
                res = requests.put(
                    url=resumable_uri,
                    headers={
                        "Content-Length": f"{end_byte - start_byte + 1}",
                        "Content-Range": f"bytes {start_byte}-{end_byte}/{file_size}",
                    },
                    data=chunk,
                )
                res.raise_for_status()

                start_byte += chunk_size

                make_request(
                    path=self.uri,
                    method="PATCH",
                    payload={"percentCompleted": (end_byte + 1) / file_size * 100},
                )
                retry_count = 0
            except Exception as e:
                if retry_count > 20:
                    print(
                        "A network error occurred. Upload failed after too many retries."
                    )

                    self.properties = make_request(
                        path=self.uri,
                        method="PATCH",
                        payload={
                            "errorMessage": "A network error occurred. Upload failed after too many retries."
                        },
                    )

                    raise e

                retry_count += 1
                time.sleep(retry_count)
                print(e)
                print("An error occurred. Retrying last chunk of resumable upload")
                start_byte = retry_partial_upload(
                    file_size=file_size, resumable_uri=resumable_uri
                )

        return

    def list_variables(self):
        # TODO
        return


def retry_partial_upload(*, retry_count=0, file_size, resumable_uri):
    print("Attempting to resume upload")

    try:
        res = requests.put(
            url=resumable_uri,
            headers={"Content-Range": f"bytes */{file_size}"},
        )

        if res.status_code == 404:
            return 0

        res.raise_for_status()

        if res.status_code == 200 or res.status_code == 201:
            return file_size
        elif res.status_code == 308:
            range_header = res.headers["Range"]

            if range_header:
                match = re.match(r"bytes=0-(\d+)", range_header)
                if match.group(0) and not math.isnan(int(match.group(1))):
                    return int(match.group(1)) + 1
                else:
                    raise Exception("An unknown error occurred. Please try again.")
            # If GCS hasn't received any bytes, the header will be missing
            else:
                return 0
    except Exception as e:
        if retry_count > 10:
            raise e

        retry_partial_upload(
            retry_count=retry_count + 1,
            file_size=file_size,
            resumable_uri=resumable_uri,
        )



from .Dataset import Dataset


class Organization:
    def __init__(self, name):
        self.name = name

    def dataset(self, name, *, version="current"):
        return Dataset(name, organization=self, version=version)



import json


class Variable:
    def __init__(self, properties):
        self.properties = properties

    def __getitem__(self, key):
        return self.properties[key]

    def __str__(self):
        return json.dumps(self.properties, indent=2)



from .Table import Table
from .Query import Query
from urllib.parse import quote as quote_uri

from ..common.api_request import make_request, make_paginated_request
import json


class Dataset:
    def __init__(
        self,
        name,
        *,
        version="current",
        user=None,
        organization=None,
        properties=None,
    ):
        self.name = name
        self.version = version
        self.user = user
        self.organization = organization
        self.identifier = (
            f"{(self.organization or self.user).name}.{self.name}:{self.version}"
        )
        self.uri = f"/datasets/{quote_uri(self.identifier)}"
        self.properties = properties

    def __getitem__(self, key):
        return self.properties[key]

    def __str__(self):
        return json.dumps(self.properties, indent=2)

    def table(self, name, *, sample=False):
        return Table(name, dataset=self, sample=sample)

    def exists(self):
        try:
            make_request(method="GET", path=self.uri)
        except Exception as err:
            if err.args[0]["status"] != 404:
                raise err
            return False

        return True

    def get(self):
        self.properties = make_request(method="GET", path=self.uri)
        self.uri = self.properties["uri"]
        return self

    def create(self, *, public_access_level="none", description=None):
        if self.organization:
            path = f"/organizations/{self.organization.name}/datasets"
        else:
            path = f"/users/{self.user.name}/datasets"

        self.properties = make_request(
            method="POST",
            path=path,
            payload={
                "name": self.name,
                "publicAccessLevel": public_access_level,
                "description": description,
            },
        )
        return self

    def update(self, *, name=None, public_access_level=None, description=None):
        payload = {}
        if name:
            payload["name"] = name
        if public_access_level:
            payload["publicAccessLevel"] = public_access_level
        if description is not None:
            payload["description"] = description

        self.properties = make_request(
            method="POST",
            path=self.uri,
            payload=payload,
        )
        return self

    def delete(self):
        make_request(
            method="DELETE",
            path=self.uri,
        )
        return

    def create_next_version(self, ignore_if_exists=False):
        if not self.properties or not hasattr(self.properties, "nextVersion"):
            self.get()

        if not self.properties["nextVersion"]:
            make_request(method="POST", path=f"{self.uri}/versions")
        elif not ignore_if_exists:
            raise Exception(
                f"Next version already exists at {self.properties['nextVersion']['datasetUri']}. To avoid this error, set argument ignore_if_exists to True"
            )

        return Dataset(
            name=self.name,
            user=self.user,
            organization=self.organization,
            version="next",
        ).get()

    def release(self):
        make_request(
            method="POST",
            path=f"{self.uri}/versions/next/release",
        )
        return Dataset(
            name=self.name,
            user=self.user,
            organization=self.organization,
            version="current",
        ).get()

    def list_tables(self, max_results):
        tables = make_paginated_request(
            path=f"{self.uri}/tables", page_size=100, max_results=max_results
        )
        return [
            Table(table["name"], dataset=self, properties=table) for table in tables
        ]

    def query(self, query):
        return Query(query, default_dataset=self.identifier)



from .Query import Query
from .Table import Table
from urllib.parse import quote as quote_uri
from ..common.api_request import make_paginated_request


class Project:
    def __init__(self, name, *, user, properties=None):
        self.user = user
        self.name = name
        self.identifier = f"{self.user.name}.{self.name}"
        self.uri = f"/projects/{quote_uri(self.identifier)}"
        self.properties = properties

    def query(self, query):
        return Query(query, default_project=self.identifier)

    def table(self, name):
        return Table(name, project=self)

    def list_tables(self, *, max_results=None, include_dataset_tables=False):
        tables = make_paginated_request(
            path=f"{self.uri}/tables",
            page_size=100,
            max_results=max_results,
            query={"includeDatasetTables": include_dataset_tables},
        )
        return [
            Table(table["name"], project=self, properties=table) for table in tables
        ]



from .Dataset import Dataset
from .Project import Project


class User:
    def __init__(self, name):
        self.name = name

    def dataset(self, name, *, version="current"):
        return Dataset(name, user=self, version=version)

    def project(self, name):
        return Project(name, user=self)



import os

try:
    import pkg_resources

    pkg_resources.declare_namespace(__name__)
except ImportError:
    import pkgutil

    __path__ = pkgutil.extend_path(__path__, __name__)

from .classes.Query import Query as _Query
from .classes.User import User as _User
from .classes.Organization import Organization as _Organization


def organization(name):
    return _Organization(name)


def user(name):
    return _User(name)


def query(
    query,
    *,
    default_dataset=os.getenv("REDIVIS_DEFAULT_DATASET"),
    default_project=os.getenv("REDIVIS_DEFAULT_PROJECT")
):
    return _Query(
        query,
        default_dataset=default_dataset,
        default_project=default_project,
    )


def table(name):
    if os.getenv("REDIVIS_DEFAULT_PROJECT") is not None:
        return (
            _User(os.getenv("REDIVIS_DEFAULT_PROJECT").split(".")[0])
            .project(os.getenv("REDIVIS_DEFAULT_PROJECT").split(".")[1])
            .table(name)
        )
    elif os.getenv("REDIVIS_DEFAULT_DATASET") is not None:
        return (
            _User(os.getenv("REDIVIS_DEFAULT_DATASET").split(".")[0])
            .dataset(os.getenv("REDIVIS_DEFAULT_DATASET").split(".")[1])
            .table(name)
        )
    else:
        raise Exception(
            "Cannot reference an unqualified table if the neither the REDIVIS_DEFAULT_PROJECT or REDIVIS_DEFAULT_DATASET environment variables are set."
        )


__all__ = ["organization", "user", "query", "table"]
